module SSG.Parser.Util

import SSG.Parser.Core

import Data.List1
import Data.Vect
import Data.String
import Control.Eff

||| Type alias for parsing errors
public export
ParseError : Type
ParseError = String

||| Type alias for a parser equipped with exception handling
public export
PS : Type -> Type
PS a = Eff [Parser, Except ParseError] a

||| Run a parser and extact the parsed value
export
runPS : PS a -> (str : String) -> Either String a
runPS x str =
  fst . extract . runParser str . runExcept $ x

||| Run a parser, extracting the parsed value, and throwing an error if the parse was not
||| complete or if it failed
export
runPS' : PS a -> (str : String) -> PS a
runPS' x str = do
  (res, rest) <- lift . runParser str . runExcept $ x
  unless (null rest) $ throw "Incomplete parse"
  rethrow res

||| Attempt to run a fallible parser, backtracking on failure
export
try : (f : PS a) -> (err : ParseError -> PS a) -> PS a
try f err = do
  state <- save
  result <- lift . runExcept $ f
  case result of
    Left e => do
      load state
      err e
    Right result => pure result

||| Variant of `try` that wraps the result in a `Maybe`
export
tryMaybe : (f : PS a) -> PS (Maybe a)
tryMaybe f = try (map Just f) (\_ => pure Nothing)

||| Variant of `try` that wraps the result in an `Either`
export
tryEither : (f : PS a) -> PS (Either ParseError a)
tryEither f = try (map Right f) (\e => pure $ Left e)

||| Attempt to parse each of a list of parsers, returning the first that
||| succeeds
export
oneOfE : (err : ParseError) -> List (PS a) -> PS a
oneOfE err [] = throw err
oneOfE err (x :: xs) = do
  res <- tryMaybe x
  case res of
    Nothing => oneOfE err xs
    Just y => pure y

||| Attempt to parse 0+ of a thing
export
many : (f : PS a) -> PS (List a)
many f = do
  Just next <- tryMaybe f
    | _ => pure []
  map (next ::) (many f)

||| Attempt to parse at least one of a thing, giving the supplied error on
||| failure
export
atLeastOne : (err : ParseError) -> (f : PS a) -> PS (List1 a)
atLeastOne err f = do
  Just head <- tryMaybe f
    | _ => throw err
  tail <- many f
  pure $ head ::: tail

||| Attempt to parse an exact string
export
thisString : String -> PS String
thisString str = do
  res <- parseExact str
  case res of
    Got x => pure x
    Failed x => throw "Expected \{show str} Got \{show x}"
    EoF => throw "Unexpected EoF"

||| Attempt to parse one of a list of strings
export
theseStrings : List String -> PS String
theseStrings strs = theseStrings' strs
  where
    theseStrings' : List String -> PS String
    theseStrings' [] = throw "Expected one of \{show strs}"
    theseStrings' (x :: xs) = do
      res <- parseExact x
      case res of
        Got y => pure y
        _ => theseStrings' xs

||| Attempt to parse one of a list of chars
export
theseChars : List Char -> PS Char
theseChars cs = do
  Just [next] <- peek 1
    | _ => throw "Unexpected EoF"
  if any (== next) cs
    then do
      _ <- parseChar (const True)
      pure next
    else throw "Expected one of \{show cs}, got \{show next}"

||| Attempt to parse any char not in the list
export
notTheseChars : List Char -> PS Char
notTheseChars cs = do
  Just [next] <- peek 1
    | _ => throw "Unexpected EoF"
  if any (== next) cs
    then throw "Expected not one of \{show cs}, got \{show next}"
    else do
      _ <- parseChar (const True)
      pure next

||| Parse an exact match, but 'replace' it with the provided value
export
exactReplace : (str : String) -> (replacement : a) -> PS a
exactReplace str replacement = do
  _ <- parseExact str
  pure replacement

||| Any character
export
any : PS Char
any = do
  x <- parseChar (const True)
  case x of
    Got y => pure y
    (Failed y) => throw "What?"
    EoF => throw "Unexpected EoF"
